local HonorTracker = select(2, ...)
local Brackets = HonorTracker:RegisterModule("Brackets")
local Stats = HonorTracker:GetModule("Stats")
local Comms = HonorTracker:GetModule("Comms")
local L = HonorTracker.L
local alreadyScanned = {}

Brackets.bracketToPoints = {0, 400} -- Rank points for each bracket
Brackets.pointsToRanks = {0, 2000} -- Rank points for each rank

for i=3, 14 do
    Brackets.bracketToPoints[i] = (i - 2) * 1000
    Brackets.pointsToRanks[i] = (i - 2) * 5000
end

function Brackets:OnLoad()
    self:RefreshPlayerData()

    -- Otherwise check for HonorSpy carry over
    if( not IsAddOnLoaded("HonorSpy") ) then return end
    if( self.globalDB.notifiedHonorSpy and self.globalDB.notifiedHonorSpy[GetRealmName()] ) then return end

    self.globalDB.notifiedHonorSpy = self.globalDB.notifiedHonorSpy or {}
    self.globalDB.notifiedHonorSpy[GetRealmName()] = true
    self.realmBracketDB.resetAt = HonorSpy.db.factionrealm.last_reset

    -- Filter out any data older than 3 days, as a rough estimation of whether it's fresh or not.
    local maxAge = 3600 * 24 * 3
    local carriedOver = 0
    for playerName, data in pairs(HonorSpy.db.factionrealm.currentStandings) do
        local existingData = self.realmBracketDB.players[playerName]
        local age = (GetServerTime() - data.last_checked)
        if( (not existingData or existingData.lastChecked < data.last_checked) and age < maxAge ) then
            carriedOver = carriedOver + 1

            self.realmBracketDB.players[playerName] = {
                class = data.class,
                lastSent = GetServerTime(),
                lastChecked = data.last_checked,
                rankPoints = data.RP,
                rankProgress = data.rankProgress,
                rank = data.rank,
                thisWeek = {
                    honor = data.thisWeekHonor,
                    kills = 15, -- We know HonorSpy only syncs on >15 kills
                },
                lastWeek = {
                    standing = data.standing,
                    honor = data.lastWeekHonor,
                }
            }        
        end
    end

    HonorTracker:Print(L["HonorSpy was detected. HonorTracker is able to do the equivalent more efficiently now, and it's recommended you disable HonorSpy."])
    HonorTracker:Print(string.format(L["We carried over %d players over to HonorTracker, and HonorTracker will sync with other HonorTracker and HonorSpy users automatically."], carriedOver), true)
end

function Brackets:OnWeeklyReset()
    self.realmBracketDB.players = {}
    self.realmBracketDB.resetAt = GetServerTime()
end

-- Save our personal data and send it off to others
function Brackets:OnDailyReset()
    -- Reset whether we inspected somebody today
    for _, data in pairs(self.realmBracketDB.players) do
        data.seenToday = nil
    end

    self:RefreshPlayerData()
end

function Brackets:RefreshPlayerData()
    local name = UnitName("player")

    local rank = select(2, GetPVPRankInfo(UnitPVPRank("player")))
    -- Calculate rank progress.
    local rankProgress = GetPVPRankProgress()
    -- Figure out last weeks info, which is what we really care about.
    local lastWeekHonor, lastWeekStanding = select(3, GetPVPLastWeekStats())
    -- This should be zero, but I don't trust Blizzard so will be paranoid.
    local thisWeekKills, thisWeekHonor = GetPVPThisWeekStats()
    
    -- Don't save our own data until we see 15 kills
    if( thisWeekKills < 15 ) then return end

    self.realmBracketDB.players[name] = {
        class = select(2, UnitClass("player")),
        seenToday = true,
        lastChecked = GetServerTime(),
        rankPoints = Stats:CalculateRankPoints(rank, rankProgress),
        rankProgress = rankProgress,
        rank = rank,
        thisWeek = {
            honor = thisWeekHonor,
            kills = thisWeekKills,
        },
        lastWeek = {
            standing = lastWeekStanding,
            honor = lastWeekHonor,
        }
    }

    Comms:SendPlayerData(name)
end

-- Estimate rank progression data
function Brackets:Estimate(playerName)
    local data = self.realmBracketDB.players[playerName]
    if( not data ) then return end
   
    -- Figure out where the players current standing is
    local currentStanding = 0
    for _, checkData in pairs(self.realmBracketDB.players) do
        if( checkData.thisWeek.honor >= data.thisWeek.honor ) then
            currentStanding = currentStanding + 1
        end
    end

    -- Calculate bracket and bracket progress
    local bracketSizes = self:CalculateBrackets()
    
    local bracket = 1
    local innerBracketProgress = 0;
    for i=2, 14 do
        -- If our estimated standing is higher than the bracket, we've found out where we are
        if( currentStanding > bracketSizes[i] ) then
            innerBracketProgress = (bracketSizes[i - 1] - currentStanding) / (bracketSizes[i - 1] - bracketSizes[i])
            break
        end

        bracket = i
    end

    -- If you're standing 1 and bracket 14, you will always be 100%
    if( bracket == 14 and currentStanding == 1) then
        innerBracketProgress = 1
    end

    -- Figure out the awarded amount of points
    local rankAward = self.bracketToPoints[bracket] + 1000 * innerBracketProgress
    -- Now factor in diminishing returns
    local estimatedRankPoints = math.floor((data.rankPoints * 0.8) + rankAward + 0.5)

    return {
        rankPoints = estimatedRankPoints,
        standing = currentStanding,
        bracket = bracket
    }
end

-- Calculate the players in each bracket
function Brackets:CalculateBrackets()
    local poolSize = 0
    for name, data in pairs(self.realmBracketDB.players) do
        poolSize = poolSize + 1
    end
    
    -- Brackets 1 -> 14
    local brackets = {1, 0.845, 0.697, 0.566, 0.436, 0.327, 0.228, 0.159, 0.100, 0.060, 0.035, 0.020, 0.008, 0.003} 
    for i=1, 14 do
        brackets[i] = math.floor((brackets[i] * poolSize))
    end

    return brackets
end

function Brackets:ScanEligibleUnit(unit)
    -- Skip in combat
    if( InCombatLockdown() ) then return end
    -- Skip in a battlefield
    if( select(2, IsInInstance()) == "pvp" ) then return end

    -- Skip if we have an inspect queued
    if( self.queuedInspectData or self.queuedHonorInspect ) then return end

    -- Skip if we've already scanned them today
    local name = UnitName(unit)
    if( self.realmBracketDB.players[name] and self.realmBracketDB.players[name].seenToday ) then
        return
    -- In case we don't record the person, filter it out locally this way too
    elseif( alreadyScanned[name] ) then
        return
    end

    -- Skip if generally not eligible
    if( UnitRealmRelationship(unit) ~= 1) then return end
    if( not UnitIsPlayer(unit) ) then return end
    if( not UnitIsFriend("player", unit) ) then return end
    if( not CheckInteractDistance(unit, 1) ) then return end
    if( not CanInspect(unit) ) then return end

    alreadyScanned[name] = true

    NotifyInspect(unit)
    RequestInspectHonorData()
end

function Brackets:PLAYER_TARGET_CHANGED()
    self:ScanEligibleUnit("target")
end

function Brackets:UPDATE_MOUSEOVER_UNIT()
    self:ScanEligibleUnit("mouseover")
end

function Brackets:INSPECT_HONOR_UPDATE()
    if( not self.queuedHonorInspect or not self.queuedHonorInspect.name ) then
        return
    elseif( self.skipHonorEvent ) then
        self.skipHonorEvent = nil
        return
    end
    
    -- Attempt to filter out inspects which look potentially buggy
    local inspectTag = string.join(",", GetInspectHonorData())
    if( self.lastSeenInspect == inspectTag ) then
        return
    end

    local todayHK, todayDK, yesterdayHK, yesterdayHonor, thisWeekHK, thisWeekHonor, lastWeekHK, lastWeekHonor, lastWeekStanding, lifetimeHK, lifetimeDK, lifetimeHighestRank = GetInspectHonorData()
    -- Filter out people who are not eligible yet
    if( thisWeekHK < 15 ) then
        self.queuedHonorInspect = nil
        self.lastSeenInspect = inspectTag
        return
    end

    local rankProgress = GetInspectPVPRankProgress()
    local rankPoints = Stats:CalculateRankPoints(self.queuedHonorInspect.rank, rankProgress)

    local existingData = self.realmBracketDB.players[self.queuedHonorInspect.name]
    self.realmBracketDB.players[self.queuedHonorInspect.name] = {
        class = self.queuedHonorInspect.class,
        seenToday = true,
        lastSent = existingData and existingData.lastSent,
        lastChecked = GetServerTime(),
        rankPoints = rankPoints,
        rankProgress = rankProgress,
        rank = self.queuedHonorInspect.rank,
        thisWeek = {
            honor = thisWeekHonor,
            kills = thisWeekHK,
        },
        lastWeek = {
            standing = lastWeekStanding,
            honor = lastWeekHonor,
        }
    }

    Comms:SendPlayerData(self.queuedHonorInspect.name)
    
    self.queuedHonorInspect = nil
    self.lastSeenInspect = inspectTag
end

hooksecurefunc("RequestInspectHonorData", function() 
    Brackets.queuedHonorInspect = Brackets.queuedInspectData
    Brackets.queuedInspectData = nil
    Brackets.skipHonorEvent = true
end)

hooksecurefunc("NotifyInspect", function(unit)
    -- Ignore cross realm inspect
    if( UnitRealmRelationship(unit) ~= 1) then
        Brackets.queuedInspectData = {}
        return
    end

    Brackets.queuedInspectData = {
        name = UnitName(unit),
        class = select(2, UnitClass(unit)),
        rank = select(2, GetPVPRankInfo(UnitPVPRank(unit))),
    }
end)

hooksecurefunc("ClearInspectPlayer", function()
    Brackets.queuedInspectData = nil
    Brackets.queuedHonorInspect = nil
end)

local eventFrame = CreateFrame("Frame")
eventFrame:RegisterEvent("INSPECT_HONOR_UPDATE")
eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
eventFrame:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
eventFrame:SetScript("OnEvent", function(self, event, ...)
    Brackets[event](Brackets, ...)
end)
